/*
 * Common keypad client functionality for input event generation.
 *
 * Copyright (C) 2018, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/input-keypad-client.h>

#include <pthread.h>
#include <pthread-l4.h>

#include <l4/re/env.h>

#include <l4/re/event_enums.h>
#include <l4/sys/thread.h>
#include <l4/util/util.h>

#include <ipc/cap_alloc.h>
#include <ipc/mem_ipc.h>

#include <stdlib.h>

#include "keypad_client.h"



/* Obtain a reference to the keypad. */

void Input_keypad_client::init_keypad()
{
  _keypad_cap = l4re_env_get_cap("keypad");
}

/* Access the keypad server memory. */

void Input_keypad_client::init_keypad_data()
{
  /* Obtain a reference to the keypad data. */

  client_Keypad keypad(_keypad_cap);

  if (keypad.get_keypad_data(&_mem))
    return;

  /* Attach the keypad data to a region in this task. */

  unsigned long size;

  if (ipc_dataspace_size(_mem, &size))
    return;

  if (ipc_attach_dataspace(_mem, size, &_keymem))
    return;

  /* Allocate memory for a copy of the keypad data. */

  _keymem_previous = malloc(size);
}

/* Release capabilities and the memory for a copy of the keypad data. */

void Input_keypad_client::release_keypad_data()
{
  if (l4_is_valid_cap(_mem))
    ipc_cap_free(_mem);

  if (l4_is_valid_cap(_keypad_cap))
    ipc_cap_free(_keypad_cap);

  if (_keymem_previous && (_keymem_previous != NULL))
    free(_keymem_previous);
}

/* Scan the keypad, compare old and new states, and generate key events. */

void Input_keypad_client::scan_keypad()
{
  uint32_t *keymem = (uint32_t *) _keymem;
  uint32_t *keymem_previous = (uint32_t *) _keymem_previous;
  int row, column, key_pressed;
  uint32_t changed, mask;
  Input_event event;

  for (column = 0; column < _keypad->columns(); column++)
  {
    /* Test for differences within each column. */

    changed = keymem[column] ^ keymem_previous[column];

    /* Transfer the current values to the previous values once used. */

    keymem_previous[column] = keymem[column];

    /* Move to the next column if no changes occurred. */

    if (!changed) continue;

    /* Inspect each changed row position. */

    mask = 1 << (_keypad->rows() - 1);

    for (row = 0; row < _keypad->rows(); row++)
    {
      if (changed & mask)
      {
        key_pressed = keymem[column] & mask;

        /* Construct an event using the appropriate key code. */

        event = (Input_event) {
          L4RE_EV_KEY, _keypad->code(column, row), key_pressed
          };

        /* Invoke the supplied handler. */

        _handler(event, _priv);
      }
      mask >>= 1;
    }
  }
}

/* Perform keypad scanning in a separate thread. */

void *Input_keypad_client::scan_mainloop(void *data)
{
  /* Since this is a static method, obtain instance details from the data. */

  Input_keypad_client *self = reinterpret_cast<Input_keypad_client *>(data);

  while (1)
  {
    if (self->_handler) self->scan_keypad();
    l4_sleep(20); /* 20ms -> 50Hz */
  }

  return 0;
}

/* Thread initialisation for the keypad scanning activity. */

void Input_keypad_client::attach(Input_handler handler, void *priv)
{
  pthread_attr_t thread_attr;
  struct sched_param sp;

  /* Record the handler and bundled private data. */

  _handler = handler;
  _priv = priv;

  /* Thread initialisation boilerplate. */

  if (pthread_attr_init(&thread_attr))
    return;

  sp.sched_priority = 0x20;
  pthread_attr_setschedparam(&thread_attr, &sp);
  pthread_attr_setschedpolicy(&thread_attr, SCHED_L4);
  pthread_attr_setinheritsched(&thread_attr, PTHREAD_EXPLICIT_SCHED);

  /* Provide this instance as the private data. */

  pthread_create(&_pthread, &thread_attr, scan_mainloop, this);
}
